'use strict';

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const fs = require('./utils/fs');
const Resolver = require('./Resolver');
const Parser = require('./Parser');
const WorkerFarm = require('./WorkerFarm');
const worker = require('./utils/promisify')(require('./worker.js'));
const Path = require('path');
const Bundle = require('./Bundle');

var _require = require('chokidar');

const FSWatcher = _require.FSWatcher;

const FSCache = require('./FSCache');
const HMRServer = require('./HMRServer');
const Server = require('./Server');

var _require2 = require('events');

const EventEmitter = _require2.EventEmitter;

const Logger = require('./Logger');
const PackagerRegistry = require('./packagers');

/**
 * The Bundler is the main entry point. It resolves and loads assets,
 * creates the bundle tree, and manages the worker farm, cache, and file watcher.
 */
class Bundler extends EventEmitter {
  constructor(main, options = {}) {
    super();
    this.mainFile = Path.resolve(main || '');
    this.options = this.normalizeOptions(options);

    this.resolver = new Resolver(this.options);
    this.parser = new Parser(this.options);
    this.packagers = new PackagerRegistry();
    this.cache = this.options.cache ? new FSCache(this.options) : null;
    this.logger = new Logger(this.options);
    this.delegate = options.delegate || {};

    this.pending = true;
    this.loadedAssets = new Map();
    this.farm = null;
    this.watcher = null;
    this.hmr = null;
    this.bundleHashes = null;
    this.errored = false;
  }

  normalizeOptions(options) {
    const isProduction = options.production || process.env.NODE_ENV === 'production';
    const publicURL = options.publicURL || '/' + Path.basename(options.outDir || 'dist');
    const watch = typeof options.watch === 'boolean' ? options.watch : !isProduction;
    return {
      outDir: Path.resolve(options.outDir || 'dist'),
      publicURL: publicURL,
      watch: watch,
      cache: typeof options.cache === 'boolean' ? options.cache : true,
      killWorkers: typeof options.killWorkers === 'boolean' ? options.killWorkers : true,
      minify: typeof options.minify === 'boolean' ? options.minify : isProduction,
      hmr: typeof options.hmr === 'boolean' ? options.hmr : watch,
      logLevel: typeof options.logLevel === 'number' ? options.logLevel : 3
    };
  }

  addAssetType(extension, path) {
    if (typeof path !== 'string') {
      throw new Error('Asset type should be a module path.');
    }

    if (this.farm) {
      throw new Error('Asset types must be added before bundling.');
    }

    this.parser.registerExtension(extension, path);
  }

  addPackager(type, packager) {
    if (this.farm) {
      throw new Error('Packagers must be added before bundling.');
    }

    this.packagers.add(type, packager);
  }

  bundle() {
    var _this = this;

    return _asyncToGenerator(function* () {
      _this.options.extensions = _this.parser.extensions;
      _this.farm = WorkerFarm.getShared(_this.options);

      if (_this.options.watch) {
        _this.watcher = new FSWatcher();
        _this.watcher.on('change', _this.onChange.bind(_this));
      }

      if (_this.options.hmr) {
        _this.hmr = new HMRServer();
        _this.options.hmrPort = _this.hmr.port;
      }

      _this.logger.status('⏳', 'Building...');

      try {
        yield fs.mkdirp(_this.options.outDir);

        _this.mainAsset = yield _this.resolveAsset(_this.mainFile);
        return yield _this.buildAsset(_this.mainAsset, true);
      } catch (err) {
        _this.errored = true;
        _this.logger.error(err);
      } finally {
        if (!_this.watcher && _this.options.killWorkers) {
          _this.stop();
        }
      }
    })();
  }

  stop() {
    if (this.farm) {
      this.farm.end();
    }

    if (this.watcher) {
      this.watcher.close();
    }

    if (this.hmr) {
      this.hmr.stop();
    }
  }

  buildAsset(asset, isInitialBundle = false) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      let startTime = Date.now();
      _this2.pending = true;
      _this2.errored = false;

      // Invalidate the asset, unless this is the initial bundle
      if (!isInitialBundle) {
        asset.invalidate();
        if (_this2.cache) {
          _this2.cache.invalidate(asset.name);
        }
      }

      // Load the asset, and its dependencies
      yield _this2.loadAsset(asset);

      // Emit an HMR update for any new assets (that don't have a parent bundle yet)
      // plus the asset that actually changed.
      if (_this2.hmr && !isInitialBundle) {
        let assets = [..._this2.findOrphanAssets(), asset];
        _this2.hmr.emitUpdate(assets);
      }

      // Invalidate bundles
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = _this2.loadedAssets.values()[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          let asset = _step.value;

          asset.invalidateBundle();
        }

        // Create a new bundle tree and package everything up.
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      let bundle = _this2.createBundleTree(_this2.mainAsset);
      _this2.bundleHashes = yield bundle.package(_this2, _this2.bundleHashes);

      // Unload any orphaned assets
      _this2.unloadOrphanedAssets();

      _this2.pending = false;
      _this2.emit('bundled', bundle);

      let buildTime = Date.now() - startTime;
      let time = buildTime < 1000 ? `${buildTime}ms` : `${(buildTime / 1000).toFixed(2)}s`;
      _this2.logger.status('✨', `Built in ${time}.`, 'green');

      return bundle;
    })();
  }

  resolveAsset(name, parent) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      var _ref = yield _this3.resolver.resolve(name, parent);

      let path = _ref.path,
          pkg = _ref.pkg;

      if (_this3.loadedAssets.has(path)) {
        return _this3.loadedAssets.get(path);
      }

      let asset = _this3.parser.getAsset(path, pkg, _this3.options);
      _this3.loadedAssets.set(path, asset);

      if (_this3.watcher) {
        _this3.watcher.add(path);
      }

      return asset;
    })();
  }

  resolveDep(asset, dep) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      try {
        return yield _this4.resolveAsset(dep.name, asset.name);
      } catch (err) {
        err.message = `Cannot resolve dependency '${dep.name}'`;

        // Generate a code frame where the dependency was used
        if (dep.loc) {
          yield asset.loadIfNeeded();
          err.loc = dep.loc;
          err = asset.generateErrorMessage(err);
        }

        err.fileName = asset.name;
        throw err;
      }
    })();
  }

  loadAsset(asset) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      if (asset.processed) {
        return;
      }

      if (!_this5.errored) {
        _this5.logger.status('⏳', `Building ${asset.basename}...`);
      }

      // Mark the asset processed so we don't load it twice
      asset.processed = true;

      // First try the cache, otherwise load and compile in the background
      let processed = _this5.cache && (yield _this5.cache.read(asset.name));
      if (!processed) {
        processed = yield _this5.farm.run(asset.name, asset.package, _this5.options);
        if (_this5.cache) {
          _this5.cache.write(asset.name, processed);
        }
      }

      asset.generated = processed.generated;
      asset.hash = processed.hash;

      // Call the delegate to get implicit dependencies
      let dependencies = processed.dependencies;
      if (_this5.delegate.getImplicitDependencies) {
        let implicitDeps = yield _this5.delegate.getImplicitDependencies(asset);
        if (implicitDeps) {
          dependencies = dependencies.concat(implicitDeps);
        }
      }

      // Process asset dependencies
      yield Promise.all(dependencies.map((() => {
        var _ref2 = _asyncToGenerator(function* (dep) {
          let assetDep = yield _this5.resolveDep(asset, dep);
          if (dep.includedInParent) {
            // This dependency is already included in the parent's generated output,
            // so no need to load it. We map the name back to the parent asset so
            // that changing it triggers a recompile of the parent.
            _this5.loadedAssets.set(dep.name, asset);
          } else {
            asset.dependencies.set(dep.name, dep);
            asset.depAssets.set(dep.name, assetDep);
            yield _this5.loadAsset(assetDep);
          }
        });

        return function (_x) {
          return _ref2.apply(this, arguments);
        };
      })()));
    })();
  }

  createBundleTree(asset, dep, bundle) {
    if (dep) {
      asset.parentDeps.add(dep);
    }

    if (asset.parentBundle) {
      // If the asset is already in a bundle, it is shared. Move it to the lowest common ancestor.
      if (asset.parentBundle !== bundle) {
        let commonBundle = bundle.findCommonAncestor(asset.parentBundle);
        if (asset.parentBundle !== commonBundle && asset.parentBundle.type === commonBundle.type) {
          this.moveAssetToBundle(asset, commonBundle);
        }
      }

      return;
    }

    // Create the root bundle if it doesn't exist
    if (!bundle) {
      bundle = new Bundle(asset.type, Path.join(this.options.outDir, asset.generateBundleName(true)));
      bundle.entryAsset = asset;
    }

    // Create a new bundle for dynamic imports
    if (dep && dep.dynamic) {
      bundle = bundle.createChildBundle(asset.type, Path.join(this.options.outDir, asset.generateBundleName()));
      bundle.entryAsset = asset;
    }

    // Add the asset to the bundle of the asset's type
    bundle.getSiblingBundle(asset.type).addAsset(asset);

    // If the asset generated a representation for the parent bundle type, also add it there
    if (asset.generated[bundle.type] != null) {
      bundle.addAsset(asset);
    }

    asset.parentBundle = bundle;

    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
      for (var _iterator2 = asset.dependencies.values()[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
        let dep = _step2.value;

        let assetDep = asset.depAssets.get(dep.name);
        this.createBundleTree(assetDep, dep, bundle);
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return) {
          _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }

    return bundle;
  }

  moveAssetToBundle(asset, commonBundle) {
    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = Array.from(asset.bundles)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        let bundle = _step3.value;

        bundle.removeAsset(asset);
        commonBundle.getSiblingBundle(bundle.type).addAsset(asset);
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3.return) {
          _iterator3.return();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }

    let oldBundle = asset.parentBundle;
    asset.parentBundle = commonBundle;

    // Move all dependencies as well
    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
      for (var _iterator4 = asset.depAssets.values()[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
        let child = _step4.value;

        if (child.parentBundle === oldBundle) {
          this.moveAssetToBundle(child, commonBundle);
        }
      }
    } catch (err) {
      _didIteratorError4 = true;
      _iteratorError4 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion4 && _iterator4.return) {
          _iterator4.return();
        }
      } finally {
        if (_didIteratorError4) {
          throw _iteratorError4;
        }
      }
    }
  }

  *findOrphanAssets() {
    var _iteratorNormalCompletion5 = true;
    var _didIteratorError5 = false;
    var _iteratorError5 = undefined;

    try {
      for (var _iterator5 = this.loadedAssets.values()[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
        let asset = _step5.value;

        if (!asset.parentBundle) {
          yield asset;
        }
      }
    } catch (err) {
      _didIteratorError5 = true;
      _iteratorError5 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion5 && _iterator5.return) {
          _iterator5.return();
        }
      } finally {
        if (_didIteratorError5) {
          throw _iteratorError5;
        }
      }
    }
  }

  unloadOrphanedAssets() {
    var _iteratorNormalCompletion6 = true;
    var _didIteratorError6 = false;
    var _iteratorError6 = undefined;

    try {
      for (var _iterator6 = this.findOrphanAssets()[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
        let asset = _step6.value;

        this.unloadAsset(asset);
      }
    } catch (err) {
      _didIteratorError6 = true;
      _iteratorError6 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion6 && _iterator6.return) {
          _iterator6.return();
        }
      } finally {
        if (_didIteratorError6) {
          throw _iteratorError6;
        }
      }
    }
  }

  unloadAsset(asset) {
    this.loadedAssets.delete(asset.name);
    if (this.watcher) {
      this.watcher.unwatch(asset.name);
    }
  }

  onChange(path) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      let asset = _this6.loadedAssets.get(path);
      if (!asset) {
        return;
      }

      _this6.logger.clear();
      _this6.logger.status('⏳', `Rebuilding ${path}...`);

      try {
        yield _this6.buildAsset(asset);
      } catch (err) {
        _this6.errored = true;
        _this6.logger.error(err);
      }
    })();
  }

  middleware() {
    return Server.middleware(this);
  }

  serve(port = 1234) {
    this.logger.persistent('Server running at ' + this.logger.chalk.cyan(`http://localhost:${port}`));
    this.bundle();
    return Server.serve(this, port);
  }
}

module.exports = Bundler;